# tests/test_eds_to_rjn_daemon_runner.py
import pytest

@pytest.mark.slow
def test_run_hourly_tabular_trend_eds_to_rjn_test_mode():
    """
    This test runs the ETL logic in test mode (no live RJN transmission).
    It will raise errors if anything in the pipeline fails.
    """
    from workspaces.eds_to_rjn.scripts.daemon_runner import run_hourly_tabular_trend_eds_to_rjn
    try:
        run_hourly_tabular_trend_eds_to_rjn(test=True)
    except Exception as e:
        pytest.fail(f"ETL run in test mode failed: {e}")




# pipeline/config_via_web.py
from __future__ import annotations # Delays annotation evaluation, allowing modern 3.10+ type syntax and forward references in older Python versions 3.8 and 3.9
import threading
import urllib.parse 
from typing import Any

from pipeline_eds.server.web_utils import launch_browser, is_server_running
from pipeline_eds.server.config_server import run_config_server_in_thread # Import the getter
from pipeline_eds.cli import GLOBAL_SHUTDOWN_EVENT
# Import the launch function (Assuming you can import it or pass it)
# Assuming 'launch_server_for_web_gui' is available in the module that imports config_via_web
# For this example, let's assume you pass the launch function to the manager if needed.


def browser_get_input(manager: Any, key: str, prompt_message: str, hide_input: bool) -> str | None:
    """
    Launches a modal/page on the MAIN server and polls for the result.
    This function now BLOCKS the calling thread (SecurityAndConfig).
    """
    global GLOBAL_SHUTDOWN_EVENT

    # 1. Register the prompt request
    try:
        request_id = manager.register_prompt(key, prompt_message, hide_input)
    except RuntimeError as e:
        # typer not imported in config_via_web, use print/log
        print(f"⚠️ Error: {e}") 
        return None 
    
    CONFIG_SERVER_URL_CHECK = manager.get_server_url()
    
    # 1. Server Status Check
    server_thread = None
    if not is_server_running(CONFIG_SERVER_URL_CHECK):
        print(f"--- Server not running. Launching temporary config server. ---")
        server_thread = run_config_server_in_thread(port=8083)

       # Now the manager has the correct URL (e.g., http://127.0.0.1:8083)
        CONFIG_SERVER_URL = manager.get_server_url()
    else:
        # Server is running, we just need to ensure the client is polling.
        # If the client isn't polling, we should launch the config modal page
        CONFIG_SERVER_URL = CONFIG_SERVER_URL_CHECK
        print(f"--- Config Server is active. Using existing server at {CONFIG_SERVER_URL} ---")

    # URL-encode the variables you want to pass
    encoded_message = urllib.parse.quote_plus(prompt_message)
    encoded_hide_input = str(hide_input) # "True" or "False" string
    #CONFIG_MODAL_URL = f"{CONFIG_SERVER_URL}/config_modal?request_id={request_id}"
    # Construct the full URL, including the new parameters
    CONFIG_MODAL_URL = (
        f"{CONFIG_SERVER_URL}/config_modal?"
        f"request_id={request_id}&"
        f"message={encoded_message}&"
        f"hide_input={encoded_hide_input}"
    )
    launch_browser(CONFIG_MODAL_URL)
    print(f"--- Config Server is active. Launched browser to dedicated modal page... ---")
    # 4. Poll for the result (Blocking)
    try:
        while True:
            # 1. Check if the main process wants to shut down
            if GLOBAL_SHUTDOWN_EVENT.is_set():
                print("\nPolling loop detected shutdown signal. Exiting gracefully.")
                # Clean up the registered prompt before exiting
                manager.clear_result(request_id)
                return None # Exit the function
            value = manager.get_and_clear_result(request_id)
            if value is not None:
                print("--- Input Received from Main UI! ---")
                return value
            #time.sleep(0.5)
            # Use a threading Event to make the sleep interruptible by signals
            threading.Event().wait(0.5)
    except KeyboardInterrupt: # 
        print("KeyboardInterrupt")
        # Clean up the registered prompt before exiting
        print("CALL manager.clear_result(request_id)")
        manager.clear_result(request_id)
        print("\nPolling cancelled by user. Returning None.")
        # Optional: You might need to stop the server_thread gracefully here if it was started
        return None
    except Exception as e: 
        # Handle server connection loss gracefully during poll
        print(f"\nPolling error: {e}")
        return None


"""
Migrated manually from eds.py by Clayton on 1 December 2025
Moving these means that the references must be changed for the currently running code. 
However, this should become a non-issue once we call the EDS SOAP API instead of the REST API or the local db files. 

We would like to move away from this and just use the SOAP api.
"""
from __future__ import annotations # Delays annotation evaluation, allowing modern 3.10+ type syntax and forward references in older Python versions 3.8 and 3.9
import logging
import typer
import pyhabitat as ph

from pipeline_eds.api.eds.rest.client import ClientEdsRest
from pipeline_eds.decorators import log_function_call

logger = logging.getLogger(__name__)
if ph.on_windows():
    import mysql.connector
else:
    pass


def _get_eds_local_db_credentials(service_name = "pipeline-eds-local-database",item_name = "eds_dbs") -> dict:
        return {}
    
def access_database_files_locally(
    session_key: str,
    starttime: int,
    endtime: int,
    point: list[int],
    tables: list[str] | None = None
) -> list[list[dict]]:

    from pipeline_eds.api.eds.rest.alarm import decode_stat
    """
    Access MariaDB data directly by querying all MyISAM tables with .MYD files
    modified in the given time window, filtering by sensor ids in 'point'.

    If 'tables' is provided, only query those tables; otherwise fall back to most recent table.

    Returns a list (per sensor id) of dicts with keys 'ts', 'value', 'quality'.

    This is provided as a fallback if API access fails.
    """

    logger.info("Accessing MariaDB directly — local SQL mode enabled.")
    workspace_name = 'eds_to_rjn'
    workspace_manager = WorkspaceManager(workspace_name)

    local_database_dict = ClientEdsRest._get_eds_local_db_credentials(service_name = "pipeline-eds-local-database",item_name = "eds_dbs")
    if not isinstance(local_database_dict,dict) or len(local_database_dict):
        typer.echo("Please develop _get_eds_local_db_credentials() to return a JSON-like dict structure, " \
        "after drawing database credentials from the keyring and compiling them into a dictionary. " \
        "And then, make the function defunct, " \
        "by implementing prompt or loading of each " \
        "secure credentialed string to at the point of sale, " \
        "with clear documentation but not and intermediate helper-funciton. " \
        "In this way, " \
        "we avoid spaghetti code and tie the demand for " \
        "information closely to the source for information." \
        "Implement the argument 'forget', "
        "if you do not want the value saved to the plaintext config file or " \
        "the cryptography-secure store credentials. ")
    secrets_dict = SecretConfig.load_config(secrets_file_path=workspace_manager.get_secrets_file_path())
    #full_config = secrets_dict["eds_dbs"][session_key]
    #conn_config = {k: v for k, v in full_config.items() if k != "storage_path"}
    
    conn_config = secrets_dict["eds_dbs"][session_key]
    results = []

    try:
        logger.info("Attempting: mysql.connector.connect(**conn_config)")
        conn = mysql.connector.connect(**conn_config)
        cursor = conn.cursor(dictionary=True)

        # Determine which tables to query
        if tables is None:
            most_recent_table = get_most_recent_table(cursor, session_key.lower())
            if not most_recent_table:
                logger.warning("No recent tables found.")
                return [[] for _ in point]
            tables_to_query = [most_recent_table]
        else:
            tables_to_query = tables

        for table_name in tables_to_query:
            if not table_has_ts_column(conn, table_name, db_type="mysql"):
                logger.warning(f"Skipping table '{table_name}': no 'ts' column.")
                continue

            for point_id in point:
                #logger.info(f"Querying table {table_name} for sensor id {point_id}")
                query = f"""
                    SELECT ts, ids, tss, stat, val FROM `{table_name}`
                    WHERE ts BETWEEN %s AND %s AND ids = %s
                    ORDER BY ts ASC
                """
                cursor.execute(query, (starttime, endtime, point_id))
                full_rows = []
                for row in cursor:
                    quality_flags = decode_stat(row["stat"])
                    quality_code = quality_flags[0][2] if quality_flags else "N"
                    full_rows.append({
                        "ts": row["ts"],
                        "value": row["val"],
                        "quality": quality_code,
                    })
                full_rows.sort(key=lambda x: x["ts"])
                results.append(full_rows)

    except mysql.connector.errors.DatabaseError as db_err:
        if "Can't connect to MySQL server" in str(db_err):
            logger.error("Local database access failed: Please run this code on the proper EDS server where the local MariaDB is accessible.")
            # Optionally:
            print("ERROR: This code must be run on the proper EDS server for local database access to work.")
            return [[] for _ in point]  # return list of empty lists, one per point
        else:
            raise  # re-raise other DB errors
    except Exception as e:
        logger.error(f"Unexpected error accessing local database: {e}")
        # hitting this in termux
        raise
    finally:
        # cleanup cursor/connection if they exist
        try:
            cursor.close()
            conn.close()
        except Exception:
            pass

    logger.info(f"Successfully retrieved data for {len(point)} point(s)")
    return results

#def identify_relevant_MyISM_tables(session_key: str, starttime: int, endtime: int, secrets_dict: dict) -> list:
# 3.8-safe, no hints
def identify_relevant_MyISM_tables(session_key, starttime, endtime, secrets_dict):
    #
    # Use the secrets file to control where your database can be found
    try:
        storage_dir = secrets_dict["eds_dbs"][str(session_key+"-config")]["storage_path"]
    except:
        logging.warning(f"User the secrets.yaml file to set the local database folder. Something like, storage_path: 'E:/SQLData/wwtf/'")
        return []
    # Collect matching table names based on file mtime
    matching_tables = []

    if False:
        for fname in os.listdir(storage_dir):
            fpath = os.path.join(storage_dir, fname)
            if not os.path.isfile(fpath):
                continue
            mtime = os.path.getmtime(fpath)
            if starttime <= mtime <= endtime:
                table_name, _ = os.path.splitext(fname)
                if 'pla' in table_name: 
                    matching_tables.append(table_name)

    '''
    # Instead of os.path.join + isfile + getmtime every time...
    # Use `os.scandir`, which gives all of that in one go and is much faster:
    with os.scandir(storage_dir) as it:
        for entry in it:
            if entry.is_file():
                mtime = entry.stat().st_mtime
                if starttime <= mtime <= endtime and 'pla' in entry.name:
                    table_name, _ = os.path.splitext(entry.name)
                    matching_tables.append(table_name)
    '''
    # Efficient, sorted, filtered scan
    sorted_entries = sorted(
        (entry for entry in os.scandir(storage_dir) if entry.is_file()),
        key=lambda e: e.stat().st_mtime,
        reverse=True
    )

    for entry in sorted_entries:
        mtime = entry.stat().st_mtime
        if starttime <= mtime <= endtime and 'pla' in entry.name:
            table_name, _ = os.path.splitext(entry.name)
            matching_tables.append(table_name)


    #print("Matching tables:", matching_tables)
    return matching_tables

def identify_relevant_tables(session_key, starttime, endtime, secrets_dict):
    try:
        conn_config = secrets_dict["eds_dbs"][session_key]
        conn = mysql.connector.connect(**conn_config)
        cursor = conn.cursor(dictionary=True)
        # Use INFORMATION_SCHEMA instead of filesystem
        #return get_ten_most_recent_tables(cursor, conn_config["database"])
        return get_n_most_recent_tables(cursor, conn_config["database"], n=80)
    except mysql.connector.Error:
        logger.warning("Falling back to filesystem scan — DB not accessible.")
        return identify_relevant_MyISM_tables(session_key, starttime, endtime, secrets_dict)

def get_most_recent_table(cursor, db_name, prefix='pla_'):
    query = f"""
        SELECT TABLE_NAME
        FROM INFORMATION_SCHEMA.TABLES
        WHERE TABLE_SCHEMA = %s AND TABLE_NAME LIKE %s
        ORDER BY TABLE_NAME DESC
        LIMIT 1;
    """
    cursor.execute(query, (db_name, f'{prefix}%'))
    result = cursor.fetchone()
    return result['TABLE_NAME'] if result else None

#def get_ten_most_recent_tables(cursor, db_name, prefix='pla_') -> list[str]:
def get_ten_most_recent_tables(cursor, db_name, prefix='pla_'):
    """
    Get the 10 most recent tables with the given prefix.
    Returns a LIST OF STRINGS, not a single string.
    """
    query = f"""
        SELECT TABLE_NAME
        FROM INFORMATION_SCHEMA.TABLES
        WHERE TABLE_SCHEMA = %s AND TABLE_NAME LIKE %s
        ORDER BY TABLE_NAME DESC
        LIMIT 10;
    """
    cursor.execute(query, (db_name, f'{prefix}%'))
    results = cursor.fetchall()
    
    # Extract table names as individual strings
    table_names = [result['TABLE_NAME'] for result in results]
    
    logger.info(f"Found {len(table_names)} recent tables with prefix '{prefix}': {table_names}")
    return table_names  # This is a LIST of strings: ['pla_68a98310', 'pla_68a97500', ...]


def get_n_most_recent_tables(cursor, db_name, n, prefix='pla_'):
    """
    Get the 10 most recent tables with the given prefix.
    Returns a LIST OF STRINGS, not a single string.
    """
    query = f"""
        SELECT TABLE_NAME
        FROM INFORMATION_SCHEMA.TABLES
        WHERE TABLE_SCHEMA = %s AND TABLE_NAME LIKE %s
        ORDER BY TABLE_NAME DESC
        LIMIT {n};
    """
    cursor.execute(query, (db_name, f'{prefix}%'))
    results = cursor.fetchall()
    
    # Extract table names as individual strings
    table_names = [result['TABLE_NAME'] for result in results]
    
    logger.info(f"Found {len(table_names)} recent tables with prefix '{prefix}': {table_names}")
    return table_names  # This is a LIST of strings: ['pla_68a98310', 'pla_68a97500', ...]

def this_computer_is_an_enterprise_database_server(secrets_dict: dict, session_key: str) -> bool:
    """
    Check if the current computer is an enterprise database server.
    This is determined by checking if the ip address matches the configured EDS database key.
    """
    import socket
    from urllib.parse import urlparse
    from pipeline_eds.helpers import get_lan_ip_address_of_current_machine
    # Check if the session_key exists in the secrets_dict
    url = secrets_dict["eds_apis"][session_key]["url"]
    parsed = urlparse(url)
    hostname = parsed.hostname  # Extract hostname from URL
    ip = socket.gethostbyname(hostname)
    bool_ip = (ip == get_lan_ip_address_of_current_machine())
    logger.info(f"Checking if this computer is enterprise database server: {bool_ip}")
    return bool_ip

@log_function_call(level=logging.DEBUG)
def demo_eds_local_database_access():
    from pipeline_eds.queriesmanager import QueriesManager
    from pipeline_eds.queriesmanager import load_query_rows_from_csv_files, group_queries_by_col
    #from pipeline_eds.api.eds.database import this_computer_is_an_enterprise_database_server, identify_relevant_tables, access_database_files_locally
    workspace_name = 'eds_to_rjn' # workspace_name = WorkspaceManager.identify_default_workspace_name()
    workspace_manager = WorkspaceManager(workspace_name)
    queries_manager = QueriesManager(workspace_manager)
    queries_file_path_list = workspace_manager.get_default_query_file_paths_list() # use default identified by the default-queries.toml file
    logger.debug(f"queries_file_path_list = {queries_file_path_list}")

    queries_dictlist_unfiltered = load_query_rows_from_csv_files(queries_file_path_list)
    queries_defaultdictlist_grouped_by_session_key = group_queries_by_col(queries_dictlist_unfiltered,'zd')
    secrets_dict = SecretConfig.load_config(secrets_file_path = workspace_manager.get_secrets_file_path())
    sessions_eds = {}

    # --- Prepare Stiles session_eds

    session_stiles = None # assume the EDS API session cannot be established
    sessions_eds.update({"WWTF":session_stiles})


    key_eds = "WWTF"
    session_key = key_eds
    session_eds = session_stiles
    point_list = [row['iess'] for row in queries_defaultdictlist_grouped_by_session_key.get(key_eds,[])]
    point_list_sid = [row['sid'] for row in queries_defaultdictlist_grouped_by_session_key.get(key_eds,[])]

    logger.info(f"point_list = {point_list}")
    # Discern the time range to use
    starttime = queries_manager.get_most_recent_successful_timestamp(api_id="WWTF")
    logger.info(f"queries_manager.get_most_recent_successful_timestamp(), key = {'WWTF'}")
    endtime = helpers.get_now_time_rounded(workspace_manager)
    starttime = TimeManager(starttime).as_unix()
    endtime = TimeManager(endtime).as_unix() 
    logger.info(f"starttime = {starttime}")
    logger.info(f"endtime = {endtime}")

    if this_computer_is_an_enterprise_database_server(secrets_dict, key_eds):
        tables = identify_relevant_tables(session_key, starttime, endtime, secrets_dict)
        results = access_database_files_locally(key_eds, starttime, endtime, point=point_list_sid, tables=tables)
    else:
        logger.warning("This computer is not an enterprise database server. Local database access will not work.")
        results = [[] for _ in point_list]
    print(f"len(results) = {len(results)}")
    print(f"len(results[0]) = {len(results[0])}")
    print(f"len(results[1]) = {len(results[1])}")
    
    for idx, iess in enumerate(point_list):
        if results[idx]:
            #print(f"rows = {rows}")
            timestamps = []
            values = []
            
            for row in results[idx]:
                #print(f"row = {row}")
                #ClientEdsRest.print_point_info_row(row)

                dt = datetime.fromtimestamp(row["ts"])
                timestamp_str = helpers.round_datetime_to_nearest_past_five_minutes(dt).isoformat(timespec='seconds')
                if row['quality'] == 'G':
                    timestamps.append(timestamp_str)
                    values.append(round(row["value"],5)) # unrounded values fail to post
            print(f"final row = {row}")
        else:
            print("No data rows for this point")

    
def table_has_ts_column(conn, table_name, db_type="mysql"):
    if db_type == "sqlite":
        with conn.cursor() as cur:
            # your sqlite logic here
            cur.execute(f"PRAGMA table_info({table_name});")
            return any(row[1] == "ts" for row in cur.fetchall())
        pass
    elif db_type == "mysql":
        with conn.cursor() as cur:
            cur.execute(f"SHOW COLUMNS FROM `{table_name}` LIKE 'ts'")
            result = cur.fetchall()
            return len(result) > 0
    else:
        raise ValueError(f"Unsupported database type: {db_type}")





import matplotlib.pyplot as plt
import numpy as np

# Sample data
x = np.linspace(0, 10, 100)
y1 = np.sin(x)
y2 = np.exp(x / 5)

# Create the first plot
fig, ax1 = plt.subplots()
ax1.plot(x, y1, 'b-', label='Sine Wave')
ax1.set_xlabel('X-axis')
ax1.set_ylabel('Sine', color='b')
ax1.tick_params(axis='y', labelcolor='b')

# Create the second y-axis
ax2 = ax1.twinx()
ax2.plot(x, y2, 'r--', label='Exponential')
ax2.set_ylabel('Exponential', color='r')
ax2.tick_params(axis='y', labelcolor='r')

# Add a title and show the plot
plt.title('Second Y-Axis Example')
plt.show()


from __future__ import annotations # Delays annotation evaluation, allowing modern 3.10+ type syntax and forward references in older Python versions 3.8 and 3.9
import os
import json
import logging.config


def setup_logging(config_file="config/logging.json"): # relative to root
    if not os.path.exists("logs"):
        os.makedirs("logs")
    with open(config_file, 'r') as f:
        config = json.load(f)
        logging.config.dictConfig(config) 

# Dynamically set logging level, per handler, like when using a GUI
#def set_handler_level(handler_name: str, level_name: str):
def set_handler_level(handler_name, level_name):
    logger = logging.getLogger()  # root logger
    level = getattr(logging, level_name.upper(), None)
    if level is None:
        raise ValueError(f"Invalid log level: {level_name}")
    
    for handler in logger.handlers:
        if handler.get_name() == handler_name:
            handler.setLevel(level)
            logging.info(f"Set {handler_name} handler level to {level_name}")
            break
    else:
        raise ValueError(f"Handler '{handler_name}' not found")
    
# Custom JSON formatter
class JSONFormatter(logging.Formatter):
    def format(self, record):
        log_record = {
            "time": self.formatTime(record, "%Y-%m-%d %H:%M:%S"),
            "level": record.levelname,
            "message": record.getMessage(),
        }
        return json.dumps(log_record, indent=4)
    
class PrettyJSONFormatter(logging.Formatter):
    def format(self, record):
        if isinstance(record.msg, (dict, list)):
            try:
                record.msg = json.dumps(record.msg, indent=2, ensure_ascii=False)
            except Exception:
                pass  # fallback to default formatting
        return super().format(record)


from __future__ import annotations
import os
import sys
import platform
from pathlib import Path
import subprocess
import datetime
from pyhabitat import on_windows

from pipeline_eds.version_info import get_package_name, get_package_version

# Importing winreg is necessary for proper Windows registry access.
# We wrap it in a try-except block for environments where it might not exist (e.g., development on Linux).
try:
    import winreg
except ImportError:
    winreg = None
    
# Constants
APP_NAME = get_package_name()
PACKAGE_NAME = get_package_name() # Used for executable name and AppData folder
INSTALL_VERSION = get_package_version()
PACKAGE_ALIAS_EXE = f"{get_package_name()}-exe" # alias for non-pipx binary
PACKAGE_ALIAS_PYZ = f"{get_package_name()}-pyz"
INSTALL_VERSION_FILE = "install_version.txt"
LOG_FILE = "install_log.txt" # New constant for logging

# --- Logging Helpers ---

def _get_log_path() -> Path:
    """Returns the full path to the installation log file."""
    return setup_appdata_dir() / LOG_FILE

def log_message(message: str, is_error: bool = False):
    """Writes a timestamped message to the installation log file."""
    timestamp = datetime.datetime.now().strftime("[%Y-%m-%d %H:%M:%S]")
    log_line = f"{timestamp} {'[ERROR]' if is_error else '[INFO]'}: {message}"
    
    # We must try to get the log path, but cannot rely on setup_appdata_dir being
    # fully successful or being called before setup_appdata_dir is complete.
    try:
        log_path = _get_log_path()
        with open(log_path, 'a', encoding='utf-8') as f:
            f.write(log_line + '\n')
    except Exception:
        # Fallback to console print if logging to file fails (e.g., permissions)
        if is_error:
            # Print errors to the console, even during silent operation
            print(f"FATAL LOGGING ERROR: {log_line}", file=sys.stderr)
        else:
            # We skip printing non-error messages to the console here to ensure silence
            pass

# --- Environment and Path Functions ---

def on_windows() -> bool:
    """Checks if the current operating system is Windows."""
    return platform.system() == "Windows"

def get_executable_path() -> Path | None:
    """
    Returns the path to the running executable (e.g., the PyInstaller .exe or shiv .pyz).
    
    Returns None if the application is running as a Python script to prevent setup from running with a source path.
    """
    if not on_windows():
        return None
    try:
        # sys.argv[0] is the path to the currently running entry point
        running_path = Path(sys.argv[0]).resolve()
        
        suffix = running_path.suffix.lower()
        
        # Only allow paths that explicitly look like deployed executables:
        # .exe (PyInstaller, or pipx shim) or .pyz (Shiv)
        if suffix in ['.exe', '.pyz']:
            return running_path
            
        # Reject anything else, including .py files or generic interpreter paths.
        return None
    except IndexError:
        return None

def setup_appdata_dir() -> Path:
    """
    Ensures the application's configuration and data directory exists in AppData/Local.
    
    Returns the path to the configuration directory.
    """
    # Use environment variable for robustness
    if not on_windows():
        return
    local_appdata = os.environ.get('LOCALAPPDATA')
    if not local_appdata:
        # Fallback using Path.home()
        config_dir = Path.home() / "AppData" / "Local" / PACKAGE_NAME
    else:
        config_dir = Path(local_appdata) / PACKAGE_NAME
        
    try:
        config_dir.mkdir(parents=True, exist_ok=True)
        # We don't log success here, as this function is called repeatedly.
    except Exception as e:
        # Use simple print as logging system relies on this dir existing
        print(f"Warning: Failed to create AppData directory {config_dir}: {e}", file=sys.stderr)
        
    return config_dir

def check_if_installed(exe_path: Path) -> bool:
    """
    Checks if the current application version has already performed setup.
    Installation is tied to the unique executable path.
    """
    config_dir = setup_appdata_dir()
    version_file = config_dir / INSTALL_VERSION_FILE
    
    if not version_file.exists():
        log_message(f"Version file {INSTALL_VERSION_FILE} not found. Proceeding with installation.")
        return False

    try:
        content = version_file.read_text().splitlines()
        
        if len(content) < 2:
            log_message("Version file is corrupt/incomplete. Reinstalling artifacts.", is_error=True)
            return False

        current_installed_version = content[0].strip()
        installed_exe_path = content[1].strip()
        
        is_same_version = current_installed_version == INSTALL_VERSION
        is_same_executable = installed_exe_path == str(exe_path)
        
        if not is_same_version:
            log_message(f"Found version {current_installed_version}, expected {INSTALL_VERSION}. Proceeding with update.", is_error=False)

        if not is_same_executable:
            log_message(f"Installed EXE path {installed_exe_path} does not match current path {exe_path}. Proceeding with update.", is_error=False)

        # We only consider it installed if both version and executable path match.
        return is_same_version and is_same_executable
        
    except Exception as e:
        log_message(f"Error reading installation version file: {e}. Reinstalling artifacts.", is_error=True)
        return False

def finalize_install_version(exe_path: Path):
    """Writes the installation marker file after a successful setup."""
    config_dir = setup_appdata_dir()
    version_file = config_dir / INSTALL_VERSION_FILE
    
    content = f"{INSTALL_VERSION}\n{exe_path}"
    
    try:
        version_file.write_text(content, encoding='utf-8')
        log_message(f"Installation version {INSTALL_VERSION} marked as installed for executable: {exe_path}")
    except Exception as e:
        log_message(f"Error writing installation version file: {e}", is_error=True)

# --- Setup Dispatcher ---

def setup_windows_integration():
    """
    Main dispatcher for all Windows-specific setup tasks.
    
    This function should be called during the first run of the application 
    on a Windows system. It now includes a version check to prevent running 
    on every startup, and logs verbose output instead of printing it.
    """
    if not on_windows():
        return

    # 1. Ensure AppData is set up first, so we have a log file location
    config_dir = setup_appdata_dir()
    
    # 2. Get the executable path
    exe_path = get_executable_path()
    if not exe_path:
        # Use simple print for this critical error, as installation won't proceed
        #print("Error: Could not determine running executable path (likely running from source). Aborting Windows setup.", file=sys.stderr)
        return
    else:
        print("pipeline_eds.windows_setup.setup_windows_integration() ...")
    
    short_path_ref = fr"%LOCALAPPDATA%\{PACKAGE_NAME}"

    # 3. Check if already installed
    if check_if_installed(exe_path):
        # Print a concise status message instead of the verbose setup prints
        print(f"[{APP_NAME}] {short_path_ref} is set up (v{INSTALL_VERSION}). Skipping setup.")
        return

    log_message(f"Starting NEW/UPDATE Windows setup for executable: {exe_path}")
    
    try:
        # Run setup tasks
        create_desktop_launcher(exe_path)
        register_context_menu(exe_path)
        register_powertoys_integration(exe_path)
        
        # Write version file only if all setup completed successfully
        finalize_install_version(exe_path)
        
        # 4. Success message to console
        print(f"[{APP_NAME}] {short_path_ref} is set up. Check log file at {config_dir / LOG_FILE} for details.")

    except Exception as e:
        log_message(f"FATAL ERROR during Windows setup: {e}", is_error=True)
        print(f"[{APP_NAME}] Setup failed. Check log file at {config_dir / LOG_FILE} for errors.", file=sys.stderr)


# --- Setup Sub-Functions (Modified to use logging) ---

def create_desktop_launcher(exe_path: Path):
    """
    Creates a simple BAT file on the user's desktop to launch the application.
    This is useful if the executable is buried deep in a build folder.
    """
    desktop_dir = Path.home() / "Desktop"
    bat_filename = f"Launch {APP_NAME}.bat"
    bat_path = desktop_dir / bat_filename

    # Simple BAT file content to execute the application
    bat_content = f"""@echo off
REM Launcher for {APP_NAME}
"{exe_path}" %*
pause
"""
    try:
        bat_path.write_text(bat_content, encoding='utf-8')
        log_message(f"Desktop launcher created: {bat_path}")
    except Exception as e:
        log_message(f"Warning: Failed to create desktop launcher {bat_path}: {e}", is_error=True)

def create_start_menu_shortcut(exe_path: Path):
    """
    Creates a simple BAT file in the user's Start Menu Programs folder.
    Windows will treat this BAT file as a Start Menu shortcut.
    """
    try:
        # Standard location for the user's Start Menu Programs folder
        start_menu_path = Path(os.environ['APPDATA']) / "Microsoft" / "Windows" / "Start Menu" / "Programs"
        start_menu_path.mkdir(parents=True, exist_ok=True)
    except Exception as e:
        log_message(f"Warning: Failed to locate or create Start Menu path: {e}", is_error=True)
        return

    bat_filename = f"{APP_NAME} Launch.bat"
    bat_path = start_menu_path / bat_filename

    # Simple BAT file content to execute the application silently
    bat_content = f"""@echo off
REM Start Menu Launcher for {APP_NAME}
"{exe_path}" %*
""" # Note: removed pause for cleaner Start Menu launch

    try:
        bat_path.write_text(bat_content, encoding='utf-8')
        log_message(f"Start Menu shortcut created: {bat_path}")
    except Exception as e:
        log_message(f"Warning: Failed to create Start Menu shortcut {bat_path}: {e}", is_error=True)



def register_context_menu(exe_path: Path):
    """
    Registers a context menu entry on folder background right-click that launches 
    a PowerShell script detailing installation and usage information.
    """
    if winreg is None:
        log_message("Warning: 'winreg' module not available. Skipping context menu setup.", is_error=True)
        return

    # 1. Determine AppData path and create PS1 file
    config_dir = setup_appdata_dir()
    ps1_filename = f"setup_info_{PACKAGE_NAME}.ps1"
    ps1_path = config_dir / ps1_filename
    
    # Determine the executable type for clarity in the PS script (e.g., EXE, PYZ)
    exe_type = exe_path.suffix.upper().lstrip('.') if exe_path.suffix else "UNKNOWN"

    # Content of the PowerShell script, including pipx info and example command
    ps1_content = f"""Write-Host "--- {APP_NAME} Installation and Usage Information ---"
Write-Host ""
Write-Host "This utility can be downloaded as a standalone PYZ, EXE, or ELF file from the GitHub Releases page:"
Write-Host "  https://github.com/City-of-Memphis-Wastewater/pipeline/releases"
Write-Host ""
Write-Host "To install the application system-wide for easy access (if pipx is installed):"
Write-Host "  pipx install {PACKAGE_NAME}"
Write-Host "Then you can run the application directly from any terminal:"
Write-Host "  {PACKAGE_NAME} trend --default-idcs"
Write-Host ""
Write-Host "Current Executable Path (Type: {exe_type}):"
Write-Host "  {exe_path}"
Write-Host "Example Execution using current file:"
Write-Host "  & '{exe_path}' trend --default-idcs"
Write-Host ""
Write-Host "The app data folder is located at: {config_dir}"
Write-Host ""
Pause
"""

    try:
        ps1_path.write_text(ps1_content, encoding='utf-8')
        log_message(f"Generated PowerShell script: {ps1_path}")
    except Exception as e:
        log_message(f"Error generating PowerShell script: {e}", is_error=True)
        return

    # 2. Define Registry paths and command
    key_path = fr"Software\Classes\Directory\Background\shell\{APP_NAME} Setup"
    command_key_path = fr"Software\Classes\Directory\Background\shell\{APP_NAME} Setup\command"
    
    command_to_run = f'powershell.exe -NoProfile -ExecutionPolicy Bypass -File "{ps1_path}"'

    try:
        # 3. Create the main key
        key = winreg.CreateKeyEx(winreg.HKEY_CURRENT_USER, key_path)
        winreg.SetValueEx(key, "", 0, winreg.REG_SZ, f"[{APP_NAME}] Info")
        winreg.CloseKey(key)

        # 4. Create the command key and set the execution string
        command_key = winreg.CreateKeyEx(winreg.HKEY_CURRENT_USER, command_key_path)
        winreg.SetValueEx(command_key, "", 0, winreg.REG_SZ, command_to_run)
        winreg.CloseKey(command_key)
        
        log_message(f"Successfully registered context menu for folder backgrounds: '{APP_NAME} Setup'")
        log_message(f"Command: {command_to_run}")

    except Exception as e:
        log_message(f"Error registering context menu: {e}", is_error=True)



def register_powertoys_integration(exe_path: Path):
    """
    Placeholder for more advanced OS-level integration (e.g., Clipboard/PowerToys).
    """
    log_message("\n--- PowerToys/Advanced Integration Note ---")
    log_message("Advanced clipboard monitoring or PowerToys integration must be handled within the application.")
    log_message("Consider registering a custom URI scheme (e.g., 'edsplot://') in the registry for deep links.")
    

# --- Cleanup Sub-Functions (Modified to use logging) ---

def cleanup_desktop_launcher():
    """Removes the desktop BAT file launcher."""
    desktop_dir = Path.home() / "Desktop"
    bat_filename = f"Launch {APP_NAME}.bat"
    bat_path = desktop_dir / bat_filename
    
    if bat_path.exists():
        try:
            bat_path.unlink()
            log_message(f"Cleaned up desktop launcher: {bat_path}")
        except Exception as e:
            log_message(f"Warning: Failed to delete desktop launcher {bat_path}: {e}", is_error=True)

def cleanup_start_menu_shortcut():
    """Removes the Start Menu BAT file launcher."""
    try:
        start_menu_path = Path(os.environ['APPDATA']) / "Microsoft" / "Windows" / "Start Menu" / "Programs"
    except KeyError:
        log_message("Warning: APPDATA environment variable not set. Skipping Start Menu cleanup.", is_error=True)
        return

    bat_filename = f"{APP_NAME} Launch.bat"
    bat_path = start_menu_path / bat_filename
    
    if bat_path.exists():
        try:
            bat_path.unlink()
            log_message(f"Cleaned up Start Menu shortcut: {bat_path}")
        except Exception as e:
            log_message(f"Warning: Failed to delete Start Menu shortcut {bat_path}: {e}", is_error=True)

def cleanup_credential_manager_keyring_entries():
    """
    Placeholder for remocing key ring items from Windows Credential Manager.
    A consistent username relevant to pipeline can be used for easy removal."
    """
    # We must define all added credentials discern which are relevant to pipeline_eds.
    try:
        from keyring.errors import PasswordDeleteError
        keyring.delete_password("service_name_placeholder", "username_placeholder")
    except PasswordDeleteError:
        pass
def cleanup_appdata_script():
    """Removes the PowerShell setup information script from AppData."""
    config_dir = setup_appdata_dir()
    ps1_filename = f"setup_info_{PACKAGE_NAME}.ps1"
    ps1_path = config_dir / ps1_filename

    if ps1_path.exists():
        try:
            # We only delete the script, leaving the main AppData folder in place
            ps1_path.unlink()
            log_message(f"Cleaned up AppData script: {ps1_path}")
        except Exception as e:
            log_message(f"Warning: Failed to delete AppData script {ps1_path}: {e}", is_error=True)

def cleanup_install_version_file():
    """Removes the installation version marker file."""
    config_dir = setup_appdata_dir()
    version_file = config_dir / INSTALL_VERSION_FILE
    
    if version_file.exists():
        try:
            version_file.unlink()
            log_message(f"Cleaned up installation version file: {version_file.name}")
        except Exception as e:
            log_message(f"Warning: Failed to delete version file {version_file}: {e}", is_error=True)
            
def cleanup_appdata_dir_if_empty():
    """
    Removes the main AppData folder if it is empty after artifact cleanup.
    """
    config_dir = setup_appdata_dir()
    try:
        # Check if the directory is empty.
        if not os.listdir(config_dir):
            config_dir.rmdir()
            log_message(f"Cleaned up empty AppData directory: {config_dir}")
        else:
            log_message(f"AppData directory {config_dir} is not empty. Leaving it in place to protect user data.")
    except FileNotFoundError:
        log_message(f"AppData directory {config_dir} not found during cleanup.", is_error=False)
    except OSError as e:
        # OSError is raised if rmdir fails because the directory is not empty or locked
        log_message(f"Warning: Could not remove AppData directory {config_dir}: {e}", is_error=True)
    except Exception as e:
        log_message(f"Warning: Failed to clean up AppData directory {config_dir}: {e}", is_error=True)
            
def cleanup_context_menu_registry():
    """Removes the context menu entries from the Windows Registry."""
    if winreg is None:
        return
        
    key_path = fr"Software\Classes\Directory\Background\shell\{APP_NAME} Setup"
    
    try:
        # Must delete the subkey (command) before the main key
        winreg.DeleteKey(winreg.HKEY_CURRENT_USER, key_path + r"\command")
        log_message(f"Cleaned up registry command subkey.")
    except FileNotFoundError:
        log_message(f"Registry command subkey not found during cleanup.")
    except Exception as e:
        log_message(f"Error cleaning up registry command subkey: {e}", is_error=True)
        
    try:
        winreg.DeleteKey(winreg.HKEY_CURRENT_USER, key_path)
        log_message(f"Cleaned up registry main key: {key_path}")
    except FileNotFoundError:
        log_message(f"Registry main key not found during cleanup.")
    except Exception as e:
        log_message(f"Error cleaning up registry main key: {e}", is_error=True)

def cleanup_windows_integration():
    """
    Performs full uninstallation cleanup of all artifacts created by 
    setup_windows_integration.
    """
    if not on_windows():
        return
        
    # We must print the start of cleanup since the main loop needs to know
    print(f"Starting Windows uninstallation cleanup for {APP_NAME}...") 
    
    cleanup_desktop_launcher()
    cleanup_start_menu_shortcut() 
    cleanup_context_menu_registry()
    cleanup_install_version_file() # Ensure this goes before attempting to remove the directory
    cleanup_appdata_script()
    cleanup_appdata_dir_if_empty()
    cleanup_credential_manager_keyring_entries()
    
    print("Windows cleanup complete.")

# Example of how this might be executed during application startup:
# if __name__ == "__main__":
#     setup_windows_integration()



from __future__ import annotations
import re
import requests

def get_point_live(session: requests.Session, iess: str) -> dict | None:
    api_url = str(session.base_url)
    query = {"filters": [{"iess": [iess], "tg": [0, 1]}]}
    resp = session.post(f"{api_url}/points/query", json=query, verify=False)
    data = resp.json()
    return data.get("points", [None])[0]

def get_points_export(session: requests.Session, filter_iess: list | None = None, zd: str | None = None) -> str:
    params = {"zd": zd or session.zd, "order": "iess"}
    if filter_iess:
        params["iess"] = ",".join(filter_iess) if isinstance(filter_iess, list) else filter_iess
    resp = session.get(f"{session.base_url}/points/export", params=params, verify=False)
    return resp.text

def get_points_metadata(session: requests.Session, iess_list: list[str]) -> dict[str, dict]:
    raw = get_points_export(session, filter_iess=iess_list)
    pattern = re.compile(r"(\w+)='([^']*)'")
    metadata = {}
    for line in raw.splitlines():
        if line.startswith("POINT "):
            attrs = dict(pattern.findall(line))
            if attrs.get("IESS") in iess_list:
                metadata[attrs["IESS"]] = attrs
    return metadata



'''
Source: https://emerson.sharepoint.com/sites/EDSSupportCostaRicaTeam/Shared%20Documents/Forms/AllItems.aspx?id=%2Fsites%2FEDSSupportCostaRicaTeam%2FShared%20Documents%2FGeneral%2FArchivos%20Clientes%2FNC%2D2501%2D6447%20%2D%2D%20Geoge%20Bennet%20%2D%2D%20Memphis%2FTT%20files&p=true&ga=1
Contact: Geovanny Martinez <Geovanny.Martinez@Emerson.com>
Ticket ID: NC-2501-6447
Date: May 14th
Modfied by: patryk.olchanowski@ttas.pl


Email Record:
    From: Martinez, Geovanny [EMR/SYSS/PWS/SJO] <Geovanny.Martinez@Emerson.com>
    Sent: Wednesday, May 14, 2025 11:00:38 AM
    To: BENNETT, GEORGE <GEORGE.BENNETT@memphistn.gov>; PWS Ovation Product Support [PROCESS/PWS/PITT] <Ovationproductsupport@Emerson.com>
    Cc: Ortega, Jose [EMR/SYSS/PWS/SJO] <jose.ortega@emerson.com>; Coronel, Luis Fernando [EMR/SYSS/PWS/SJO] <LuisFernando.Coronel@Emerson.com>
    Subject: RE: Ovation Call NC-2501-6447 Inquiry about the EDS API package [Ref:NC-2501-0886]
    Hello George,

    Based on the sent code and your comments, this is the response from our EDS Developers:

    From the code, it looks like they are trying to fetch the trend result immediately after sending the request. However, querying historical data may take any amount of time - from a few seconds to even several minutes.

    For all such requests that return a requestId, the proper approach is to check the task status using the following request periodically:

    GET /api/v1/requests?id=<requestId>

    Only once the result is ready should it be retrieved.
'''


from datetime import datetime
import time
import requests


API_URL = 'http://172.19.4.127:43084/api/v1/'
POINTS = ['M100FI.UNIT0@NET0', 'FI8001.UNIT0@NET0']
'''
In Python, variables written in all caps are typically used to represent constants. 
While Python does not enforce immutability for such variables, using all caps is a widely accepted convention to indicate that the value should not be changed during the program's execution.
Do not reassign these variables.
'''

def login_to_session():
    session = requests.Session()

    data = {'username': 'admin', 'password': '', 'type': 'rest client'} # type can be 'script' or 'rest client', it works the same.
    res = session.post(f'{API_URL}/login', json=data, verify=False).json()
    session.headers['Authorization'] = 'Bearer ' + res['sessionId']
    return session

def create_tabular_request(session):
    data = {
        'period': {
            'from': int(datetime(2024, 12, 16, 15).timestamp()),
            'till': int(datetime(2024, 12, 16, 18).timestamp()),
        },
        'step': 600,
        'items': [{
            'pointId': {'iess': p},
            'shadePriority': 'DEFAULT',
            'function': 'AVG'
        } for p in POINTS],
    }
    res = session.post(f'{API_URL}/trend/tabular', json=data, verify=False).json()
    return res['id']


def wait_for_request_execution(session, req_id):
    st = time.time()
    while True:
        time.sleep(1)
        res = session.get(f'{API_URL}/requests?id={req_id}', verify=False).json()
        status = res[str(req_id)]
        if status['status'] == 'FAILURE':
            raise RuntimeError('request [{}] failed: {}'.format(req_id, status['message']))
        elif status['status'] == 'SUCCESS':
            break
        elif status['status'] == 'EXECUTING':
            print('request [{}] progress: {:.2f}\n'.format(req_id, time.time() - st))

    print('request [{}] executed in: {:.3f} s\n'.format(req_id, time.time() - st))


def get_tabular(session, req_id):
    results = [[] for _ in range(len(POINTS))]
    while True:
        response = session.get(f'{API_URL}/trend/tabular?id={req_id}', verify=False).json()
        for chunk in response:
            if chunk['status'] == 'TIMEOUT':
                raise RuntimeError('timeout')

            for idx, samples in enumerate(chunk['items']):
                results[idx] += samples
            
            if chunk['status'] == 'LAST':
                return results
            
def main():
    session = login_to_session()
    req_id = create_tabular_request(session)
    wait_for_request_execution(session, req_id)
    results = get_tabular(session, req_id)

    session.post(f'{API_URL}/logout', verify=False)

    for idx, iess in enumerate(POINTS):
        print('\n{} samples:'.format(iess))
        for s in results[idx]:
            print('{} {} {}'.format(datetime.fromtimestamp(s[0]), s[1], s[2]))
if __name__ == "__main__":
    main()


from __future__ import annotations

class EdsAPIError(RuntimeError):
    """Base exception for all EDS API errors"""
    pass

class EdsTimeoutError(EdsAPIError, ConnectionError):
    """Raised when EDS server is unreachable (no VPN, timeout)"""
    pass

class EdsAuthError(EdsAPIError, PermissionError):
    """Raised when login fails due to bad credentials"""
    pass

class EdsRequestError(EdsAPIError):
    """Raised when API returns error status but connection succeeded"""
    def __init__(self, message: str, status_code: int | None = None):
        self.status_code = status_code
        super().__init__(message)


class EdsLoginException(Exception):
    """
    Custom exception raised when a login to the EDS API fails.

    This exception is used to differentiate between a simple network timeout
    and a specific authentication or API-related login failure.
    """

    def __init__(self, message: str = "Login failed for the EDS API. Check VPN and credentials."):
        """
        Initializes the EdsLoginException with a custom message.

        Args:
            message: A descriptive message for the error.
        """
        self.message = message
        super().__init__(self.message)

    @staticmethod
    def connection_error_message(url)-> None:
        print(f"\n--- AN ERROR OCCURRED ---")
        
        print("\nPlease check:")
        print(f"1. Is the IP address {url} correct and reachable?")
        print("2. Is the EDS server running?")
        print("3. Are your username and password correct?")
        return None
    



# pipeline/api/mission_opcua.py
from __future__ import annotations
from datetime import datetime, timedelta
import csv
from io import StringIO
import time
from typing import List, Dict

from opcua import Client, ua
from opcua.crypto import security_policies
from opcua.common.uaerrors import UaError

class MissionOPCUAClient:
    """
    MissionOPCUAClient handles connection and data retrieval from the Mission OPC UA Server.
    
    📝 Notes on Authentication and Security:
    - Requires X.509 certificate for application authentication (self-signed or CA-issued).
    - Use OpenSSL or opcua tools to generate cert/key if not provided.
      Example: openssl req -x509 -newkey rsa:2048 -keyout client_key.pem -out client_cert.pem -days 365 -nodes -subj "/CN=YourClient/O=YourOrg"
    - Server certificate must be trusted; fetch it during connection if needed.
    - Username/password for user authentication (request from Mission support; may differ from web creds).
    - Historical access assumes server support (confirm via manual or support).
    - Address space: Browse to discover; typical: Objects > [Customer] > [RTUs] > [DeviceName or ID] > [DataPoints e.g., AnalogInputs > AI1 > Value]
    - Adapt node paths based on browsing your server's structure (use the browse_example method).
    - For production, register client cert with Mission if required.
    - Download OPC UA User Manual for address space details: https://www.123mc.com/wp-content/uploads/2017/04/OPC-UA-User-Manual-3.docx
    """

    def __init__(self, username: str, password: str, cert_path: str, key_path: str, endpoint: str = "opc.tcp://opcua.123mc.com:4840/"):
        self.endpoint = endpoint
        self.username = username
        self.password = password
        self.cert_path = cert_path
        self.key_path = key_path
        self.client = Client(self.endpoint)
        self.client.set_security(
            security_policies.SecurityPolicyBasic256Sha256,
            certificate_path=self.cert_path,
            private_key_path=self.key_path,
            mode=ua.MessageSecurityMode.SignAndEncrypt
        )
        self.client.application_uri = "urn:your:client:uri"  # Customize with your URI
        self.client.set_user(self.username)
        self.client.set_password(self.password)
        self.customer_id = None  # Set after browsing if needed
        self.connected = False

    def connect(self):
        try:
            self.client.connect()
            self.connected = True
            print("Connected to OPC UA Server.")
        except UaError as e:
            raise ValueError(f"Connection failed: {e}. Check certs, creds, firewall (port 4840), or contact support.")

    def disconnect(self):
        if self.connected:
            self.client.disconnect()
            self.connected = False

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def browse_example(self, start_node_id: str = "i=85"):  # i=85 is Objects folder
        """Browse and print address space structure for discovery."""
        node = self.client.get_node(ua.NodeId.from_string(start_node_id))
        print(f"Browsing from {node.get_browse_name().Name}")
        children = node.get_children()
        for child in children:
            print(f"- {child.get_browse_name().Name} (NodeId: {child.nodeid})")
            # Recurse if folder: if child.get_node_class() == ua.NodeClass.Object: self.browse_example(str(child.nodeid))

    def get_device_node(self, device_name: str, device_id: int = None):
        """Find device/RTU node; adapt based on your address space (use browse_example first)."""
        # Assume structure: Objects > Customer > RTUs > DeviceName
        objects = self.client.get_objects_node()
        # Traverse to RTUs folder (replace with actual path)
        rtus_folder = objects.get_child(["0:YourCustomer", "0:RTUs"])  # Placeholder; adapt
        device_node = None
        for child in rtus_folder.get_children():
            name = child.get_browse_name().Name
            if name == device_name or (device_id and str(device_id) in name):
                device_node = child
                break
        if not device_node:
            raise ValueError(f"Device '{device_name}' not found. Browse to confirm path.")
        return device_node

    def get_analog_nodes(self, device_node: ua.Node):
        """Find analog data nodes under device; adapt based on structure."""
        # Assume: Device > AnalogInputs > AI1 > Value, AI2 > Value, etc.
        analog_folder = device_node.get_child("0:AnalogInputs")  # Placeholder
        analog_nodes = {}
        for child in analog_folder.get_children():
            if child.get_node_class() == ua.NodeClass.Object:
                value_node = child.get_child("0:Value")  # Or "0:ScaledValue"
                analog_nodes[child.get_browse_name().Name] = value_node
        return analog_nodes

    def get_analog_table(self, device_id: int, device_name: str, start_ms: int, end_ms: int, max_values: int = 0):
        """Retrieve historical analog data (like table); assumes history support."""
        if not self.connected:
            self.connect()
        device_node = self.get_device_node(device_name, device_id)
        analog_nodes = self.get_analog_nodes(device_node)
        if not analog_nodes:
            raise ValueError("No analog nodes found.")
        
        start_time = datetime.utcfromtimestamp(start_ms / 1000)
        end_time = datetime.utcfromtimestamp(end_ms / 1000)
        results = {"analogMeasurements": []}
        
        for name, node in analog_nodes.items():
            details = ua.ReadRawModifiedDetails()
            details.IsReadModified = False
            details.StartTime = ua.uatypes.DateTime(start_time)
            details.EndTime = ua.uatypes.DateTime(end_time)
            details.NumValuesPerNode = max_values  # 0 = all
            details.ReturnBounds = False
            
            result = node.history_read(details)
            for dv in result.HistoryData.DataValues:
                results["analogMeasurements"].append({
                    "point": name,
                    "value": dv.Value.Value,
                    "timestamp": dv.SourceTimestamp.timestamp() * 1000,  # ms
                    "status": dv.StatusCode.name
                })
        
        return results

    def download_analog_csv(self, device_id: int, device_name: str, start_date: str, end_date: str, file_name: str = None):
        """Download historical analog data as CSV bytes."""
        # Parse dates (YYYYMMDD to ms)
        start_dt = datetime.strptime(start_date, "%Y%m%d")
        end_dt = datetime.strptime(end_date, "%Y%m%d") + timedelta(days=1)  # End inclusive
        start_ms = int(start_dt.timestamp() * 1000)
        end_ms = int(end_dt.timestamp() * 1000)
        
        table_data = self.get_analog_table(device_id, device_name, start_ms, end_ms)
        measurements = table_data.get("analogMeasurements", [])
        
        if file_name is None:
            file_name = f"Analog_{device_name.replace(' ', '')}_DataPoints_{start_date}.csv"
        
        output = StringIO()
        writer = csv.writer(output)
        writer.writerow(["Point", "Value", "Timestamp (ms)", "Status"])
        for meas in measurements:
            writer.writerow([meas["point"], meas["value"], meas["timestamp"], meas["status"]])
        
        return output.getvalue().encode("utf-8")

def demo_retrieve_analog_data_and_save_csv():
    from pipeline_eds.env import SecretConfig
    from pipeline_eds.workspace_manager import WorkspaceManager
    workspace_name = WorkspaceManager.identify_default_workspace_name()
    workspace_manager = WorkspaceManager(workspace_name)

    secrets_dict = SecretConfig.load_config(secrets_file_path=workspace_manager.get_secrets_file_path())
    username = secrets_dict.get("contractor_apis", {}).get("Mission", {}).get("opc_username")  # Use OPC-specific creds
    password = secrets_dict.get("contractor_apis", {}).get("Mission", {}).get("opc_password")
    cert_path = secrets_dict.get("contractor_apis", {}).get("Mission", {}).get("client_cert_path")
    key_path = secrets_dict.get("contractor_apis", {}).get("Mission", {}).get("client_key_path")

    with MissionOPCUAClient(username, password, cert_path, key_path) as client:
        # Optional: Browse to discover structure
        # client.browse_example()

        # Get last 24 hours of analog table data
        end = datetime.utcnow()
        start = end - timedelta(days=1)
        to_ms = lambda dt: int(dt.timestamp() * 1000)
        table_data = client.get_analog_table(device_id=22158, device_name="Gayoso Pump Station", start_ms=to_ms(start), end_ms=to_ms(end))
        print(f"table_data = {table_data}")
        print(f"Fetched {len(table_data.get('analogMeasurements', []))} measurements.")

        # Download CSV for 6–11 Oct 2025
        csv_bytes = client.download_analog_csv(
            device_id=22158,
            device_name="Gayoso Pump Station",
            start_date="20251006",
            end_date="20251011"
        )
        with open("Gayoso_Analog.csv", "wb") as f:
            f.write(csv_bytes)

if __name__ == "__main__":
    demo_retrieve_analog_data_and_save_csv()



# pipeline_eds.version_info.py
from __future__ import annotations # Delays annotation evaluation, allowing modern 3.10+ type syntax and forward references in older Python versions 3.8 and 3.9
import sys
import toml
from importlib.metadata import version, PackageNotFoundError
from pathlib import Path  

from pipeline_eds.system_info import SystemInfo

# -- Versioning --
PIP_PACKAGE_NAME = "pipeline-eds"

def get_version_from_known_alias():
    try:
        PIPELINE_VERSION = version(PIP_PACKAGE_NAME)
    except PackageNotFoundError:
        PIPELINE_VERSION = "0.0.0"
    return PIPELINE_VERSION

PIPELINE_VERSION = get_version_from_known_alias()
try:
    __version__ = version(PIP_PACKAGE_NAME)

except PackageNotFoundError:
    # fallback if running from source
    try:
            with open(Path(__file__).parent / "VERSION") as f:
                __version__ = f.read().strip()
    except FileNotFoundError:
        __version__ = "dev" 

# --- Version Retrieval ---

def get_package_version() -> str:
    """Read from standard [project] ."""
    data = toml.load("pyproject.toml")
    project = data.get("project", {})
    return project.get("version", "0.0.0")

def get_package_name() -> str:
    """Read from standard [project] ."""
    data = toml.load("pyproject.toml")
    project = data.get("project", {})
    return project.get("name", "pipeline_eds")


def get_python_version():
    py_major = sys.version_info.major
    py_minor = sys.version_info.minor
    py_version = f"py{py_major}{py_minor}"
    return py_version

def form_dynamic_binary_name(package_name: str, package_version: str, py_version: str, os_tag: str, arch: str) -> str:    
    # Use hyphens for the CLI/EXE/ELF name
    return f"{package_name}-{package_version}-{py_version}-{os_tag}-{arch}"

if __name__ == "__main__":
    package_name = get_package_name()
    package_version = get_package_version()
    py_version = get_python_version()
    
    sysinfo = SystemInfo()
    os_tag = sysinfo.get_os_tag()
    architecture = sysinfo.get_arch()

    bin_name = form_dynamic_binary_name(package_name, package_version, py_version, os_tag, architecture)
    print(f"bin_name = {bin_name}")


# tests/test_sys_path_debug.py
from pipeline_eds import helpers, philosophy  # Make sure at least one real module is used

def test_show_sys_path():
    import sys
    print("\n--- sys.path ---")
    for p in sys.path:
        print(p)

    assert hasattr(helpers, "__file__")  # ✅ harmless
    helpers.function_view()
    print(philosophy.Philosophy())



from __future__ import annotations
from contextlib import contextmanager
from .session import login_to_session_with_credentials
from .exceptions import EdsTimeoutError

class ClientEdsRest:
    def __init__(self, credentials: dict):
        self.credentials = credentials
        self.session = None

    @contextmanager
    def connect(self):
        try:
            self.session = login_to_session_with_credentials(self.credentials)
            yield self.session
        except EdsTimeoutError:
            print("\n[EDS CLIENT] Connection to the EDS API timed out. Please check your VPN connection and try again.")
            raise
        finally:
            if hasattr(self, "session") and self.session:
                try:
                    self.session.close()
                except:
                    pass

    def __enter__(self):
        self.session = login_to_session_with_credentials(self.credentials)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.session:
            self.session.close()
        return False



import pygal
from datetime import datetime, timedelta
import random
from pathlib import Path


def generate_mock_data(num_points=24):
    now = datetime.now()
    times = [now - timedelta(hours=i) for i in reversed(range(num_points))]
    labels = [t.strftime("%H:%M") for t in times]

    data = {
        "Sensor A": [random.uniform(0, 10) for _ in times],
        "Sensor B": [random.uniform(2, 8) for _ in times],
    }
    return labels, data


def generate_svg_report(filename="report.svg"):
    labels, data = generate_mock_data()

    chart = pygal.Line(
        x_label_rotation=35,
        show_minor_x_labels=False,
        style=pygal.style.LightStyle,
        fill=True,
        show_legend=True,
        width=1000,
        height=400,
    )

    chart.title = "Hourly Sensor Readings"
    chart.x_labels = labels
    chart.x_labels_major = labels[::4]

    for label, values in data.items():
        chart.add(label, values)

    out_path = Path(filename).resolve()
    chart.render_to_file(out_path)
    print(f"✅ SVG report saved to: {out_path}")


if __name__ == "__main__":
    generate_svg_report()



from __future__ import annotations # Delays annotation evaluation, allowing modern 3.10+ type syntax and forward references in older Python versions 3.8 and 3.9
import os
import toml
from datetime import datetime
import json
import csv
from collections import defaultdict
import logging

from pipeline_eds import helpers
from pipeline_eds.time_manager import TimeManager

logger = logging.getLogger(__name__)
'''
Goal:
Set up to use the most recent query:
use-most-recently-edited-query-file = true # while true, this will ignore the files variable list and instead use a single list of the most recent files
'''

class QueriesManager:
    #def __init__(self, workspace_manager: object):
    def __init__(self, workspace_manager):
        self.workspace_manager = workspace_manager
        logger.info(f"QueriesManager using project: {self.workspace_manager.workspace_name}")
        if not workspace_manager:
            raise ValueError("workspace_manager must be provided and not None.")
        self.workspace_manager = workspace_manager

    
    def load_tracking(self):
        file_path = self.workspace_manager.get_timestamp_success_file_path()
        try:
            #logger.info({"Trying to load tracking file at": file_path})
            logger.debug({
                "event": "Loading tracking file",
                "path": str(file_path)
            })
            data = helpers.load_json(file_path)
            #logger.info({"Tracking data loaded": data})
            logger.debug({
                "event": "Tracking data loaded",
                "data": data
            })

            return data
        except FileNotFoundError:
            return {}
        
    def save_tracking(self,data):
        file_path = self.workspace_manager.get_timestamp_success_file_path()
        with open(file_path, 'w') as f:
            json.dump(data, f, indent=2)
    
    def get_most_recent_successful_timestamp(self, api_id):# -> int:
        print("QueriesManager.get_most_recent_successful_timestamp()")
        from pipeline_eds.helpers import load_toml
        config = load_toml(self.workspace_manager.get_configuration_file_path())
        try:
            timezone_config = config["settings"]["timezone"] ## configuration-example
        except:
            timezone_config = "America/Chicago"
            
        data = self.load_tracking()
        
        if not data:
            # No stored value found — go back one hour from now, rounded down to nearest 5 minutes
            one_hour_ago_local = TimeManager(datetime.now()).as_unix() - 3600  # now - 1 hour in unix seconds
            one_hour_ago_local = TimeManager(one_hour_ago_local).as_datetime()
            #one_hour_ago_utc = TimeManager.from_local(one_hour_ago_local, zone_name = timezone_config)
            tm = TimeManager(one_hour_ago_local).round_down_to_nearest_five()
        else:
            # Stored value found — parse ISO timestamp and round down to nearest 5 minutes
            last_success_iso = TimeManager(data[api_id]["timestamps"]["last_success"]).as_datetime()
            #last_success_utc = TimeManager.from_local(last_success_iso, zone_name = timezone_config).as_datetime()
            tm = TimeManager(last_success_iso).round_down_to_nearest_five()
            
        return tm

    
    def update_success(self,api_id,success_time=None):
        # This should be called when data is definitely transmitted to the target API. 
        # A confirmation algorithm might be in order, like calling back the data and checking it against the original.
        data = self.load_tracking()
        if api_id not in data:
            data[api_id] = {"timestamps": {}}
        #now = success_time or datetime.now().isoformat()
        now = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        data[api_id]["timestamps"]["last_success"] = now
        data[api_id]["timestamps"]["last_attempt"] = now
        self.save_tracking(data)

    def update_attempt(self, api_id):
        data = self.load_tracking()
        if api_id not in data:
            logger.info(f"Creating new tracking entry for {api_id}")
            data[api_id] = {"timestamps": {}}
        #now = datetime.now().isoformat()
        now = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        data[api_id]["timestamps"]["last_attempt"] = now
        self.save_tracking(data)
        logger.info(f"Updated last_attempt for {api_id}: {now}")

def load_query_rows_from_csv_files(csv_paths_list):
    queries_dictlist_unfiltered = []
    for csv_path in csv_paths_list:
        with open(csv_path, newline='') as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                queries_dictlist_unfiltered.append(row)
    return queries_dictlist_unfiltered

def group_queries_by_col(queries_array,grouping_var_str='zd'):
    queries_array_grouped = defaultdict(list)
    for row in queries_array:
        row_filter = row[grouping_var_str] 
        queries_array_grouped[row_filter].append(row)
    return queries_array_grouped

if __name__ ==  "__main__":
    pass



